#undef  DEFINE_A0
#define DEFINE_A0  /*1;*/
#include "test.h"

#undef  DEFINE_A0
#define DEFINE_A0 /*2;*/
#include "test.h"

#undef  DEFINE_A0
#define DEFINE_A0 /*3;*/
#include "test.h"

#undef  DEFINE_A0
#define DEFINE_A0 /*4;*/
#include "test.h"

